#ifndef lint
static char sccsid[] = "@(#)tiff_read.c	1.12 2/19/89";
#endif

/*
 * Copyright (c) 1988 by Sam Leffler.
 * All rights reserved.
 *
 * This file is provided for unrestricted use provided that this
 * legend is included on all tape media and as a part of the
 * software program in whole or part.  Users may copy, modify or
 * distribute this file at will.
 */

/*
 * TIFF Library.
 * Scanline-oriented Read Support
 *
 * XXX: Using different calling conventions according
 *   the file format is terrible, but I'm too lazy to handle
 *   the packing/unpacking issues (for arbitrary bits/sample).
 */
#include "tiffio.h"

extern	char *malloc();

#define	ReadOK(fd, buf, size)	(read(fd, (char *)buf, size) == size)
#define	SeekOK(fd, off, whence)	(lseek(fd, (long)off, whence) == (long)off)
#define	howmany(x, y)	(((x)+((y)-1))/(y))
#define	roundup(x, y)	((((x)+((y)-1))/(y))*(y))

TIFFReadScanline(tif, va_alist)
	register TIFF *tif;
	va_dcl
{
	static char module[] = "TIFFReadScanline";
	register TIFFDirectory *td;
	int bitspersample, scanline, s; 
	va_list ap;

	if (tif->tif_mode == O_WRONLY) {
		TIFFError(module, "%s: File not open for reading",
		    tif->tif_name);
		return (-1);
	}
	td = tif->tif_dir;
	if (!TIFFFieldSet(td, FIELD_COMPRESSION)) {
		TIFFError(module, "%s: No data compression algorithm set",
		    tif->tif_name);
		return (-1);
	}
	if (++tif->tif_row >= td->td_imagelength)	/* out of range */
		return (-1);
	if ((tif->tif_flags & TIFF_BUFFERSETUP) == 0) {
		bitspersample = td->td_bitspersample;
		tif->tif_bytespersample =
		    bitspersample <= 8 ? 1 : bitspersample <= 16 ? 2 : 4;
		scanline = bitspersample * td->td_imagewidth;
		if (td->td_planarconfig == PLANARCONFIG_CONTIG)
			scanline *= td->td_samplesperpixel;
		scanline = howmany(scanline, 8);
		if (scanline > tif->tif_scanlinesize) {
			if (tif->tif_scanline) {
				free(tif->tif_scanline);
				tif->tif_scanlinesize = 0;
			}
			tif->tif_scanline = malloc(scanline);
			if (tif->tif_scanline == NULL) {
				TIFFError(module,
				    "%s: No space for scanline buffer",
				    tif->tif_name);
				return (-1);
			}
			tif->tif_scanlinesize = scanline;
		}
		if (td->td_compression == COMPRESSION_NONE) {
			tif->tif_rawdata = tif->tif_scanline;
			tif->tif_rawdatasize = tif->tif_scanlinesize;
		} else {
			tif->tif_rawdata = NULL;
			tif->tif_rawdatasize = 0;
		}
		tif->tif_rawcc = -1;
		tif->tif_datamask = (1 << td->td_bitspersample) - 1;
		tif->tif_flags |= TIFF_BUFFERSETUP;
	}

	/*
	 * Fetch data from file, decompress it, and unpack
	 * to place scanline in user supplied buffer.
	 */
	va_start(ap);
	switch (td->td_planarconfig) {
	case PLANARCONFIG_CONTIG:
		/*
		 * Refill data if raw data buffer is empty,
		 * or if we cross over to a new strip.
		 */
		if (tif->tif_rawcc <= 0 ||
		    (tif->tif_row % td->td_rowsperstrip) == 0)
			if (!TIFFReadStrip(tif))
				goto bad;
		/*
		 * Decompress data into scanline buffer
		 */
		if (!(*tif->tif_decoderow)(tif))
			goto bad;
		bcopy(tif->tif_scanline, va_arg(ap, u_char *),
		    tif->tif_scanlinesize);
		break;
	case PLANARCONFIG_SEPARATE:
		for (s = 0; s < td->td_samplesperpixel; s++) {
			/*
			 * Refill data if raw data buffer is empty,
			 * or if we cross over to a new strip.
			 */
			if (tif->tif_rawcc <= 0 ||
			    (tif->tif_row % td->td_rowsperstrip) == 0)
				if (!TIFFReadStrip(tif))
					goto bad;
			/*
			 * Decompress data into scanline buffer
			 */
			if (!(*tif->tif_decoderow)(tif))
				goto bad;
			bcopy(tif->tif_scanline, va_arg(ap, u_char *),
			    tif->tif_scanlinesize);
		}
		break;
	}
	va_end(ap);
	return (1);
bad:
	va_end(ap);
	return (-1);
}

/*
 * Read a strip of data from the file.
 * The current row is used to select
 * the strip.
 */
static
TIFFReadStrip(tif)
	TIFF *tif;
{
	static char module[] = "TIFFReadStrip";
	register struct strip *sp;
	TIFFDirectory *td = tif->tif_dir;
	int s;

	sp = &td->td_strips[tif->tif_row / td->td_rowsperstrip];
	if (!SeekOK(tif->tif_fd, sp->offset, L_SET)) {
		TIFFError(module, "%s: Seek error at scanline %d",
		    tif->tif_name, tif->tif_row);
		return (0);
	}
	/*
	 * Expand raw data buffer, if needed, to
	 * hold data strip coming from file
	 * (perhaps should set upper bound on
	 *  the size of a buffer we'll use?).
	 */
	if (sp->bytecount > tif->tif_rawdatasize) {
		if (tif->tif_rawdata) {
			/*
			 * Dumpmode points the scanline buffer pointer
			 * into the raw buffer to avoid copy -- thus
			 * we check here to invalidate the pointer.
			 */
			if (tif->tif_rawdata <= tif->tif_scanline &&
		    tif->tif_scanline < tif->tif_rawdata+tif->tif_rawdatasize)
				tif->tif_scanline = NULL;
			free(tif->tif_rawdata);
			tif->tif_rawdata = NULL;
		}
		tif->tif_rawdatasize = roundup(sp->bytecount, 1024);
		tif->tif_rawdata = malloc(tif->tif_rawdatasize);
		if (tif->tif_rawdata == NULL) {
			TIFFError(module,
			    "%s: No space for data buffer at scanline %d",
			    tif->tif_name, tif->tif_row);
			tif->tif_rawdatasize = 0;
			return (0);
		}
		if (tif->tif_scanline == NULL)
			tif->tif_scanline = tif->tif_rawdata;
	}
	if (!ReadOK(tif->tif_fd, tif->tif_rawdata, sp->bytecount)) {
		TIFFError(module, "%s: Read error at scanline %d",
		    tif->tif_name, tif->tif_row);
		return (0);
	}
	tif->tif_rawcp = tif->tif_rawdata;
	tif->tif_rawcc = sp->bytecount;
	if (tif->tif_decodestrip)
		(*tif->tif_decodestrip)(tif);
	return (1);
}
